/******************************************************************************
*
* project name:    TIGCC Tools Suite
* file name:       ebook.c
* initial date:    19/08/2000
* author:          thomas.nussbaumer@gmx.net
* description:     a generic ebook reader which uses TTArchives with special
*                  extension "ebk" as ebooks
*
* NOTE: the first entry of an ebook must be a 2-plane 160x100 pixels title
*       image with name "title" in compressed state. Otherwise the ebook will
*       be rejected.
*       The rest of the ebook content must be compressed textfiles which were
*       previously processed with ttdos2ebk or any similar tool to remove the
*       line breaks within paragraphs.
*
*
* -----------------------------------------------------------------------------
* large font versions:
* -----------------------------------------------------------------------------
* If you specify -DLARGE_FONT on the commandline for compilation the 6x8 font
* will be used to display ebooks AND on the TI92p/V200 the title screen will
* be scaled up to fullscreen.
*
*
* -----------------------------------------------------------------------------
* This program needs TIGCC 0.95 or later to compile (F-Line, comments,
* linker optimizations). F-Line ROM_CALLs are slower, but the core of the
* ebook reader doesn't use any.
* Note that I use modified TIGCC headers, so that the large (>100 bytes)
* unnecessary __need_in_use_bit patch is not inclued.
* -----------------------------------------------------------------------------
*
*
* $Id: ebook.c,v 2.06 2004/07/14 14:31:55 DEBROUX Lionel Exp $
*
******************************************************************************/

#if defined(USE_TI89) && defined(USE_TI92PLUS)
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif defined(USE_TI92PLUS)
#define C89_92(x,y) (y)
#else
#error Please specify either -DUSE_TI89 or -DUSE_TI92PLUS on commandline
#endif


#define NO_EXIT_SUPPORT      /* doesn't need exit support */
#define NO_AMS_CHECK

//#define TIME_MEASUREMENT

//-----------------------------------------------------------------------------
// necessary stuff for unpacking
//-----------------------------------------------------------------------------
#include "ttunpack.h"
#include "ttarchive.h"


//-----------------------------------------------------------------------------
// title of reader and date of build
//-----------------------------------------------------------------------------
#define READER_TITLE   "TICT eBook-Reader v2.06"
#define READER_DATE    "14/07/2004"


//-----------------------------------------------------------------------------
// program comment
//-----------------------------------------------------------------------------
#define COMMENT_STRING READER_TITLE


//-----------------------------------------------------------------------------
// top and bottom line of menu
//-----------------------------------------------------------------------------
#define MENU_TOP       READER_TITLE" "READER_DATE
#define MENU_BOTTOM    "\x15&\x16=select  enter=choose  esc=exit"


//-----------------------------------------------------------------------------
// folder definitions
//-----------------------------------------------------------------------------
#define BOOKS_FOLDER   (SYMSTR_CONST("tictebks"))
#define MARKS_FOLDER   (SYMSTR_CONST("tictmark"))
#define MARKS_FOLDER2  "tictmark"


//-----------------------------------------------------------------------------
// how many pages are allowed per part
//-----------------------------------------------------------------------------
#define MAX_PAGES   1000


//-----------------------------------------------------------------------------
// HERE COME CALCULATOR DEPENDENT DEFINES
//-----------------------------------------------------------------------------
#define WIDTH_IN_PIXELS   C89_92(160,240)
#define HEIGHT_IN_PIXELS  C89_92(100,128)
#define K_LEFT            C89_92(338,337)
#define K_RIGHT           C89_92(344,340)
#define K_UP              C89_92(337,338)
#define K_DOWN            C89_92(340,344)
#define K_OFF2            C89_92(16651,8459)


//-----------------------------------------------------------------------------
// some useful key definitions
//-----------------------------------------------------------------------------
#define K_ESCAPE    264
#define K_ENTER      13
#define K_F1      0x10c
#define K_F5      0x110
#define K_OFF      4363
#define K_SECOND   4096

//-----------------------------------------------------------------------------
// precalculated extension of a page
//-----------------------------------------------------------------------------
#if defined(LARGE_FONT)
#define LINES_PER_PAGE  ((HEIGHT_IN_PIXELS-7)/8)
#else
#define LINES_PER_PAGE  ((HEIGHT_IN_PIXELS-6)/7)
#endif


#include <tigcclib.h>


//-----------------------------------------------------------------------------
// type definition of the bookmarks structure
//
// NOTE: the part of the latest saved bookmark gets the high bit set
//-----------------------------------------------------------------------------
#define NR_BOOKMARKS 10

typedef struct {
    unsigned short  part[NR_BOOKMARKS];   // parts are numbered starting at 1
    unsigned short  page[NR_BOOKMARKS];   // pages are numbered starting at 1
} BookMarks;


#define BOOKMARKS_MASK_LATEST   0x8000
#define IS_LATEST_MARK(m)       ((m) & BOOKMARKS_MASK_LATEST)
#define REAL_MARK_VALUE(m)      ((m) & ~(BOOKMARKS_MASK_LATEST))


//-----------------------------------------------------------------------------
// globals
//-----------------------------------------------------------------------------
short          exit_completely = 0;  // flag which indicates a F5 press
unsigned char  active_book[9]  = {}; // name of active book
BookMarks*     active_marks    = 0;  // pointer to active bookmarks structure
unsigned char* screenbuffer    = 0;  // should be set to the screenbuffer to use
unsigned char* doublebuffer    = 0;  // allocated memory for double buffering
short          nr_parts        = 0;  // number of parts of a book
short          act_part        = 0;  // actual part of a book
short          load_part       = -1; // info about which part and
short          load_page       = -1; // which page we are going to load
INT_HANDLER    old_inthandler  = NULL;
volatile short restore_status  = 0;



//=============================================================================
// interrupt handler which restores statusline if wanted
//=============================================================================
DEFINE_INT_HANDLER (RestoreStatusHandler) {
    if (restore_status) {
        memcpy((void *)((unsigned long)LCD_MEM+(HEIGHT_IN_PIXELS-7)*30),doublebuffer+(HEIGHT_IN_PIXELS-7)*30,7*30);
    }
    ExecuteHandler (old_inthandler);
}



//=============================================================================
// setup charset
//=============================================================================
extern short SetupCharSet(void);
asm("
.data
	.even
	.xdef SetupCharSet
SetupCharSet:
| The method is fast in both cases. Therefore, there is no need to suppress interrupts.

	pea (%a2)
	move.l %d3,-(%sp)
	
	movea.l 0xC8.w,%a2

| A cmpi.w would be enough...
	cmpi.l #1000,-4(%a2)
	bgt.s __Prepare_fonts_AMS_2xx__
__Prepare_fonts_AMS_1xx__:
| Get the address of DrawStr.
	movea.l 0x1A9*4(%a2),%a0
	
| Check if we have an old PedroM version.
	cmpi.l #0x50656472,-(%a0) | 'Pedr', a signature of PedroM.
	bne.s __The_ROM_we_re_running_on_is_not_PedroM__

| We have PedroM.

| Setup all fonts. This will suppress crashes consecutive to the use of a
| font that was not initialized...

| d(pc) mode is forbidden for general purpose code.
"
#ifdef LARGE_FONT
"
	lea __F_6x8_Font__(%pc),%a1
	subq.l #4,%a0
	move.l -(%a0),(%a1)
	move.l -(%a0),-(%a1)
	bra.s __Prepare_fonts_end__
"
#else
"
	lea __F_4x6_Font__(%pc),%a1
	move.l -12(%a0),(%a1)
"
#endif
"
__The_ROM_we_re_running_on_is_not_PedroM__:
| Retrieving the addresses of the fonts is easy (but slightly nasty) on AMS 1.xx.

| Catch a pc-relative jump to a subroutine in which are the three addresses we're looking for.
	lea 58+4(%a0),%a0 | Get rid of the cmpi.l #0x50656472,-(%a0) above.
	movea.w (%a0),%a1
	adda.l %a0,%a1
| a1 now contains the absolute address of the subroutine.

| d(pc) mode is forbidden for general purpose code.
	lea __FontsTable__(%pc),%a0

| Setup all fonts. This will suppress crashes consecutive to the use of a
| font that was not initialized...
| Setup F_4x6
	move.l 122(%a1),(%a0)+
"
#ifdef LARGE_FONT
"
| Setup F_6x8
	move.l 106(%a1),(%a0)+
"
#endif
/*
| Setup F_8x10
	move.l 82(%a1),(%a0)
*/
"
	bra.s __Prepare_fonts_end__

.even
__Prepare_fonts_AMS_2xx__:
| The fonts can be redefined on AMS 2.xx. So, retrieving the addresses of the fonts must be
| performed with FlashApp functions.

	movea.l 0x45D*4(%a2),%a0 | EV_runningApp
	move.w (%a0),%d3
	beq.s __Prepare_fonts_GetAttr__
| OO_GetAppAttr calls OO_GetAttr with the frame of the given application.
| So we are going to catch the frame of the running application, to pass it
| to OO_CondGetAttr.
| We have AMS 2.xx, we can use ROM_CALL 441 'HeapTable' (smaller and faster
| than pushing d3.w on the stack, calling HeapDeref, correcting sp).
        movea.l 0x441*4(%a2),%a0 | HeapTable
	lsl.w #2,%d3
	movea.l 0(%a0,%d3.w),%a0

| The pFrame of an ACB structure is located 20 bytes after the beginning.
	movea.l 20(%a0),%a0
	
	bra.s _Prepare_fonts_GetAttr__

__Prepare_fonts_GetAttr__:
        lea 0xFF000000.l,%a0
_Prepare_fonts_GetAttr__:
	movea.l 0x3FA*4(%a2),%a2 | OO_CondGetAttr
| We use OO_CondGetAttr because it does not throw errors, unlike OO_GetAttr
| (which can call OO_CondGetAttr anyway).

/*
| Setup all fonts. This will suppress crashes consecutive to the use of a
| font that was not initialized...
*/
| Setup F_4x6
| d(pc) mode is forbidden for general purpose code.
	pea __F_4x6_Font__(%pc)
	pea 0x300.w
	pea (%a0)

	jsr (%a2)
	tst.w %d0
	beq.s __Prepare_fonts_problem__
"
#ifdef LARGE_FONT
"
| Setup F_6x8
	addq.l #4,8(%sp)
	addq.w #1,6(%sp)
	jsr (%a2)
	tst.w %d0
	beq.s __Prepare_fonts_problem__
"
#endif
/*
| Setup F_8x10
	addq.l #4,8(%sp)
	addq.w #1,6(%sp)
	jsr (%a2)
	tst.w %d0
	beq.s __Prepare_fonts_problem__
*/
"
	lea 12(%sp),%sp

__Prepare_fonts_end__:
	moveq #1,%d0

_Prepare_fonts_end__:
| Don't forget to restore registers...
	move.l (%sp)+,%d3
	move.l (%sp)+,%a2

| This is not an inlined version of the routine...
	rts

__Prepare_fonts_problem__:
	moveq #0,%d0
	bra.s _Prepare_fonts_end__
");


extern unsigned char* __FontsTable__;
extern unsigned char* __F_4x6_Font__;
extern unsigned char* __F_6x8_Font__;
//extern unsigned char* __F_8x10_Font__;

asm("
| The three pointers are always stored into the program. This wastes 8 bytes maximum, but
| increases the speed of some of the functions.
| The pointer to F_4x6 also points to the character widths of F_4x6, due to the way the
| characters are stored.
	.xdef __FontsTable__
	.even
__FontsTable__:
	.xdef __F_4x6_Font__
	.even
__F_4x6_Font__:
	.long 0

	.xdef __F_6x8_Font__
	.even
__F_6x8_Font__:
	.long 0
/*	
	.xdef __F_8x10_Font__
	.even
__F_8x10_Font__:
	.long 0
*/
");



//=============================================================================
// fills a specified number of lines with given byte value (NOTE: on the TI89
// it will fill more than what is visible on the screen, but we don't care)
//=============================================================================
#define FillFullLines(y,nr,c)    memset(screenbuffer+(y)*30,c,30*(nr))



//=============================================================================
// draws string using own sprites (about 4 times faster than DrawStr)
//=============================================================================
extern void __attribute__((__regparm__(3))) FastStringXY(register short x asm("%d0"),register short y asm("%d1"),register const unsigned char* s asm("%a0"));
asm("
.data
    .global FastStringXY
FastStringXY:
    movem.l  %d3/%a2-%a4,-(%sp)

| d0 = x (NTS: do not clobber !).
| d1 = y (NTS: used only once at the beginning, d1 quickly becomes scratch
| register).
| d2 = scratch register (byteoffset, character).
| a0 = string.
    add.w    %d1,%d1
    move.w   %d1,%d2
    lsl.w    #4,%d2
    sub.w    %d1,%d2
    move.l   screenbuffer(%pc),%a1
    adda.w   %d2,%a1
| a1 = screenbuffer + 30*y.

    move.l   __F_4x6_Font__(%pc),%a2
| a2 = start of font data (sprites).
    
_beginning_of_loop_FSXY_:
| Get next character.
    moveq    #0,%d2
    move.b   (%a0)+,%d2
    jbeq     _finished_FSXY_

    add.w    %d2,%d2
    move.w   %d2,%d1
    add.w    %d2,%d2
    add.w    %d1,%d2
    lea      1(%a2,%d2.w),%a4
| a4 = address of sprite data.

    move.w   %d0,%d1
    lsr.w    #4,%d1
    add.w    %d1,%d1
    lea      0(%a1,%d1.w),%a3
| a3 = address where the upper line of the sprite will be drawn (preliminary).
    
    cmpi.w   #6*0x67,%d2 | 6 * 'g'...
| Short branches are faster if not taken.
    jbeq     _have_small_g_letter_FSXY_

_draw_sprite_FSXY_:
    move.w   %d0,%d1
    andi.w   #0xF,%d1
    
    moveq    #8,%d2
    sub.w    %d1,%d2
    moveq    #5-1,%d1
    tst.w    %d2
    bge.s    _one_word_FSXY_

    neg.w    %d2

| NOTE: We are using XOR here, so this function can also be used
| for the 'white on black' strings
_loop_one_long_FSXY_:
    moveq    #0,%d3
    move.b   (%a4)+,%d3
    swap     %d3
    lsr.l    %d2,%d3
    eor.l    %d3,(%a3)
    lea      30(%a3),%a3
    
    dbf      %d1,_loop_one_long_FSXY_

    bra.s    _next_letter_FSXY_

_loop_one_word_FSXY_:
    lea      30(%a3),%a3

_one_word_FSXY_:
    clr.w    %d3
    move.b   (%a4)+,%d3
    lsl.w    %d2,%d3
    eor.w    %d3,(%a3)
    
    dbf      %d1,_loop_one_word_FSXY_

_next_letter_FSXY_:
    moveq    #0,%d1
    move.b   -6(%a4),%d1
    add.w    %d1,%d0
    jbra     _beginning_of_loop_FSXY_
    
_finished_FSXY_:
    movem.l  (%sp)+,%d3/%a2-%a4
    rts
    
_have_small_g_letter_FSXY_:
    lea      30(%a3),%a3
    jbra     _draw_sprite_FSXY_
");


#if defined(LARGE_FONT)
//=============================================================================
// draws string using own sprites (about 4 times faster than DrawStr)
// (variant with 6x8 characters)
//=============================================================================
extern void __attribute__((__regparm__(3))) BigFastStringXY(register short x asm("%d0"),register short y asm("%d1"),register const unsigned char* s asm("%a0"));
asm("
.data
    .global BigFastStringXY
BigFastStringXY:
    movem.l  %d3/%a2-%a4,-(%sp)

| d0 = x (NTS: do not clobber !).
| d1 = y (NTS: used only once at the beginning, d1 quickly becomes scratch
| register).
| d2 = scratch register (byteoffset, character).
| a0 = string.
    add.w    %d1,%d1
    move.w   %d1,%d2
    lsl.w    #4,%d2
    sub.w    %d1,%d2
    move.l   screenbuffer(%pc),%a1
    adda.w   %d2,%a1
| a1 = screenbuffer + 30*y.

    move.l   __F_6x8_Font__(%pc),%a2
| a2 = start of font data (sprites).
    
_beginning_of_loop_BFSXY_:
| Get next character.
    moveq    #0,%d2
    move.b   (%a0)+,%d2
    jbeq     _finished_BFSXY_

    lsl.w    #3,%d2
    lea      0(%a2,%d2.w),%a4
| a4 = address of sprite data.

    move.w   %d0,%d1
    lsr.w    #4,%d1
    add.w    %d1,%d1
    lea      0(%a1,%d1.w),%a3
| a3 = address where the upper line of the sprite will be drawn.
    
_draw_sprite_BFSXY_:
    move.w   %d0,%d1
    andi.w   #0xF,%d1
    
    moveq    #8,%d2
    sub.w    %d1,%d2
    moveq    #8-1,%d1
    tst.w    %d2
    bge.s    _one_word_BFSXY_

    neg.w    %d2

| NOTE: We are using XOR here, so this function can also be used
| for the 'white on black' strings
_loop_one_long_BFSXY_:
    moveq    #0,%d3
    move.b   (%a4)+,%d3
    swap     %d3
    lsr.l    %d2,%d3
    eor.l    %d3,(%a3)
    lea      30(%a3),%a3
    
    dbf      %d1,_loop_one_long_BFSXY_

    bra.s    _next_letter_BFSXY_

_loop_one_word_BFSXY_:
    lea      30(%a3),%a3

_one_word_BFSXY_:
    clr.w    %d3
    move.b   (%a4)+,%d3
    lsl.w    %d2,%d3
    eor.w    %d3,(%a3)
    
    dbf      %d1,_loop_one_word_BFSXY_

_next_letter_BFSXY_:
    addq.w   #6,%d0
    jbra     _beginning_of_loop_BFSXY_
    
_finished_BFSXY_:
    movem.l  (%sp)+,%d3/%a2-%a4
    rts
");
#endif


//=============================================================================
// calculates x position that given string is centered
//=============================================================================
short CalcCenterPosition(const char* s,short f) {
    return (WIDTH_IN_PIXELS-DrawStrWidth(s,f))>>1;
}



//=============================================================================
// draws the titlebar containing the given String
//=============================================================================
void DrawTitleBar(const char* s) {
    FillFullLines(0,7,0xff);
    FastStringXY(CalcCenterPosition(s,F_4x6),1,s);
}



//=============================================================================
// returns decompressed entry from TTArchive
//
// returns 0 or pointer to allocated memory
// the allocated size is stored in parameter *size
//
// No checking if archive is valid or index exists is done within this function
//=============================================================================
short out_of_mem = 0;
unsigned char* GetDecompressedEntry(unsigned char*  archive,
                                    unsigned short  index,
                                    unsigned short* size)
{
    unsigned char* start;
    unsigned char* buf;

    out_of_mem = 0;
    //-------------------------------------------------------------------------
    // assume that we will come here with only valid indices !!
    // so no checks anymore
    //-------------------------------------------------------------------------
    start = ttarchive_data(archive,index);

    if (!ttunpack_valid(start)) return 0;

    *size = ttunpack_size(start);

    if (!(buf = malloc(*size))) {
        out_of_mem = 1;
        return 0;
    }

    if (ttunpack_decompress(start,buf)) {
        free(buf);
        return 0;
    }

    return buf;
}



//=============================================================================
// try to find corresponding bookmark file in bookmarks folder
//=============================================================================
SYM_ENTRY* FindBookMarks(unsigned char* bookname) {
    char       buf[30];
    char*      bp = buf;

    *bp++ = 0;
    sprintf(bp,MARKS_FOLDER2"\\%s",bookname);

    return SymFindPtr(bp+strlen(bp),0);
}



//=============================================================================
// load bookmarks
//
// if the part of one bookmark has its hibit set it will
// be copied to variables load_part/load_page
//
// otherwise load_part/load_page are set to 1
// (load automatically part1/page1)
//=============================================================================
static inline void LoadBookMarks(void) {
    short          i;
    unsigned char* src;
    SYM_ENTRY*     entry = FindBookMarks(active_book);

    load_part = 1;
    load_page = 1;

    memset(active_marks,0,sizeof(BookMarks));

    if (!entry) return;
    if (!(src = HLock(entry->handle))) return;
    src+=2;
    memcpy(active_marks,src,sizeof(BookMarks));
    HeapUnlock(entry->handle);

    unsigned short *ptr = &active_marks->part[0];
    for (i=0;i<NR_BOOKMARKS;i++) {
        if (IS_LATEST_MARK(*ptr)) {
            load_part = REAL_MARK_VALUE(*ptr);
            load_page = *(ptr+10);
//            load_page = active_marks->page[i];
            if (load_part > nr_parts || !load_page) {
                load_part = 1;
                load_page = 1;
            }
            break;
        }
        ptr++;
    }
}



//=============================================================================
// save bookmarks (the return value is not queried yet)
//=============================================================================
void/*short*/ SaveBookMarks(void) {
    SYM_ENTRY*     symptr = FindBookMarks(active_book);
    unsigned short* src;
    unsigned char  buffer[30];
    unsigned char* bp = buffer;
    HANDLE         h;
    unsigned short size;

    // if there is already a bookmark file re-use it
    if (symptr) {
        if ((symptr->flags.flags_n&(SF_ARCHIVED | SF_LOCKED)) ||
             !(src = (unsigned short *)HLock(symptr->handle)) ) return;// 0;
        src++;
        memcpy((void *)src,active_marks,sizeof(BookMarks));
        HeapUnlock(symptr->handle);
//        return 1;
    }

    *bp++ = 0;

    sprintf(bp,MARKS_FOLDER2"\\%s",active_book);

    FolderAdd(MARKS_FOLDER);  // always try to generate the folder so that
                              // there will no dialog pop up

    size = strlen(bp);
    if (!(h = HeapAlloc(sizeof(BookMarks)+8))) {
        SymDel(bp+size);
        return;// 0;
    }

    if (!(symptr = DerefSym(SymAdd(bp+size)))) {
        HeapFree(h);
        return;// 0;
    }

    symptr->handle = h;

    // get bookmarks file
    if (!(src = (unsigned short *)HLock(h))) {
        SymDel(bp+size);
        return;// 0;
    }

    // calculations evaluated during compile time.
    // No optimizations necessary!!

    // Handle starts at an even address -> we can write words.
    *src++ = (((sizeof(BookMarks)+6) / 256) << 8) + ((sizeof(BookMarks)+6) % 256);

    memcpy((void *)src,active_marks,sizeof(BookMarks));
    (unsigned char *)src += sizeof(BookMarks);
    // Handle starts at an even address -> we can write words.
    *src++ = ((0 << 8) + ('c'));
    *src++ = (('f' << 8) + ('g'));
    *src++ = ((0 << 8) + (OTH_TAG));

    HeapUnlock(h);
    return;// 1;
}



//=============================================================================
// get input key pressed by user
//
// replacement for ngetchx() which handle APD and OFF key combinations
//=============================================================================
short GetUserInput(void) {
    void *kbq = kbd_queue();
    unsigned short key;

    while (1) {
        OSTimerRestart(APD_TIMER);
        //---------------------------------------------------------------------
        // wait until we get a key. if the APD timer expires -> turn
        // calculator off
        //---------------------------------------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (!GrayCheckRunning()) idle();
            else pokeIO(0x600005,0b10111);

            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
        }

        //---------------------------------------------------------------------
        // handle calculator-off key combinations. if the APD timer expires ->
        // turn calculator off
        //---------------------------------------------------------------------
        if (key == K_OFF || key == K_OFF2) {
            off();
        }
        //---------------------------------------------------------------------
        // special treatment of F5
        //---------------------------------------------------------------------
        else if (key == K_F5) {
            exit_completely = 1;
            return K_ESCAPE;
        }
        return key & 0xf7ff;
    }
}



//=============================================================================
// handles the book mark dialog
//=============================================================================
static inline short BookMarkDialog(void) {
#define sy (HEIGHT_IN_PIXELS/2 - 2)
    SCR_RECT   rect;
    short      read;
    LCD_BUFFER b;

    LCD_save(b);
    
    rect.l = (((4L << 24) + 
             ((unsigned long)(sy - 10) << 16) + 
             ((WIDTH_IN_PIXELS-5) << 8) + 
             (sy+10)) - 0x01010000UL + 0x00000101UL);

    ScrRectFill(&rect,ScrRect,A_NORMAL);

    //-------------------------------------------------------------------------
    // the following rect.l operations perform the following separate tasks,
    // but are shorter in length (note that GCC optimizes them into a single
    // instruction).
    //-------------------------------------------------------------------------
    //rect.xy.x0++;
    //rect.xy.y0++;
    //rect.xy.x1--;
    //rect.xy.y1--;
    rect.l += 0x01010000;
    rect.l -= 0x00000101;
    ScrRectFill(&rect,ScrRect,A_REVERSE);

    DrawStr(C89_92(5,45),sy-3,"Press 0-9 to set Bookmark",A_NORMAL);

    do {
        read = GetUserInput();
        if (read == 'X' || read == 'x') read = '0';
        if ((read >= '0' && read <= '9') || read == K_ESCAPE) break;
    }
    while (1);

    LCD_restore(b);
    if (read == K_ESCAPE) return -1;
    return (read - '0');
#undef sy
}



//=============================================================================
// Displays title of eBook
//
// !! index 0 of archive must be a 2-planes 160x100 image and name title
//
// returns 0 in case of error or keycode
//=============================================================================
static inline short ShowTitle(unsigned char* archive) {
    TTARCHIVE_ENTRY* tmp = ttarchive_desc(archive,0);
    unsigned char*   buf;
    unsigned short   size;
    short            ret;
#if !defined(LARGE_FONT) || defined(USE_TI89)
    short            i;
#endif
    unsigned char*   dest1;
    unsigned char*   dest2;
    unsigned char*   src1;
    unsigned char*   src2;

    if (!tmp) return 0;

    //-------------------------------------------------------------------------
    // simple check for name title
    //-------------------------------------------------------------------------
    if (strncmp("title",tmp->name,8)) return 0;

    if (!(buf = GetDecompressedEntry(archive,0,&size))) return 0;
    if (size != 4000) {
        free(buf);
        return 0;
    }


    if (!GrayOn()) {
        free(buf);
        out_of_mem = 1;
        return 0;
    }

    dest1 = GrayGetPlane(LIGHT_PLANE);
    dest2 = GrayGetPlane(DARK_PLANE);
    src1  = buf;
    src2  = buf+2000;

    memset(dest1,0,LCD_SIZE);
    memset(dest2,0,LCD_SIZE);

#if defined(USE_TI89)
    //-------------------------------------------------------------------------
    // TI89 version
    //-------------------------------------------------------------------------
    for (i=0;i<100;i++) {
        memcpy(dest1,src1,20);
        memcpy(dest2,src2,20);
        dest1 += 30;
        dest2 += 30;
        src1  += 20;
        src2  += 20;
    }

#else

    //-------------------------------------------------------------------------
    // TI92P/V200 large font version
    //-------------------------------------------------------------------------
#if defined(LARGE_FONT)

#define STPIXEL(p,x,y)   (*((((unsigned char*)(p))+(((y)<<5)-((y)<<1)+((x)>>3)))) |= (((unsigned char)(0x80 >> ((x)&7)))))
#define LDPIXEL(p,x,y)   (*((((unsigned char*)(p))+(((y)<<4)+((y)<<2)+((x)>>3)))) &  (((unsigned char)(0x80 >> ((x)&7)))))
    {
    unsigned short x,y;
    for(x=0;x<LCD_WIDTH;x++)
      for(y=0;y<LCD_HEIGHT;y++)
        {if(LDPIXEL(src1,(unsigned)(x*160)/240,(unsigned)(y*100)/128))STPIXEL(dest1,x,y);
        if(LDPIXEL(src2,(unsigned)((239-x)*160)/240,(unsigned)(y*100)/128))STPIXEL(dest2,239-x,y);}
    }

#else

    //-------------------------------------------------------------------------
    // TI92P/V200 small font version
    //-------------------------------------------------------------------------
    dest1 += 30*14 + 5;
    dest2 += 30*14 + 5;

    for (i=0;i<100;i++) {
        memcpy(dest1,src1,20);
        memcpy(dest2,src2,20);
        dest1 += 30;
        dest2 += 30;
        src1  += 20;
        src2  += 20;
    }

#endif
#endif

    ret = GetUserInput();

    free(buf);
    GrayOff();

    return ret;
}



//=============================================================================
// fetches next line from inputbuffer into outputbuffer with respect to the
// maximum available bytes of the inputbuffer (max_avail)
//
// returns number of bytes processed from inputbuffer
//
// NOTE: (1) spaces at the end of a line are skipped
//       (2) if a word doesn't end at the end of a line and there was
//           a space previously detected, this not fitting word will be
//           left for the next line. This will make reading the document
//           much easier, because words are not splitted between lines !!!
//       (3) due to the fact that not all characters have the same
//           size this routine may return more than WIDTH_IN_PIXELS/4 characters,
//           because it tries to pack as much characters as possible into
//           one line. So no space is wasted.
//=============================================================================
unsigned short NextLine(unsigned char* inputbuffer,
                        unsigned char* outputbuffer,
                        unsigned short max_avail)
{
    unsigned short processed             = 0;
    unsigned char* input_orig            = inputbuffer;
    short          line_length           = 0;
    unsigned short last_space_processed  = 0;
    unsigned char* last_space_pos        = 0; // just to make compiler happy
    unsigned char  actchar;

    do {
        actchar = *inputbuffer++;

        if (actchar == 0x0a) {   // linebreak found - close output string
            *outputbuffer++ = 0; // and "break" out of looping ...
            processed++;
            break;
        }

        //---------------------------------------------------------------------
        // if we'll find a space remember its position and how much bytes we've
        // processed so far ...
        //---------------------------------------------------------------------
        if (actchar == ' ') {
            last_space_pos       = outputbuffer;
            last_space_processed = processed;
        }

        *outputbuffer++ = actchar;
        processed++;

#if defined(LARGE_FONT)
        line_length += 6;
#else
        line_length += (((unsigned char *)__F_4x6_Font__))[(*(outputbuffer-1))*6];
#endif
    }
    while (processed != max_avail && line_length < (WIDTH_IN_PIXELS-5));

    //-------------------------------------------------------------------------
    // if we have the end of the part reached do no further processing,
    // but return immediately
    //-------------------------------------------------------------------------
    if (processed != max_avail) {
        //---------------------------------------------------------------------
        // if we are located within a word, go back to last known space if
        // there is one ...
        //---------------------------------------------------------------------
        if (last_space_processed) {
            if (actchar      != ' '  && // last processed byte was no space
                actchar      != 0x0a && // last processed byte was no linebreak
                *inputbuffer != ' ')    // NEXT byte is NO space
            {
                processed    = last_space_processed;
                outputbuffer = last_space_pos;
            }
        }

        //---------------------------------------------------------------------
        // if actual character isn't a linebreak, skip trailing spaces if there
        // are any. if it IS a linebreak DON'T SKIP spaces. The spaces after a
        // linebreak are inserted by INTENTION! (at least I think so ;-)
        //---------------------------------------------------------------------
        if (actchar != 0x0a) {
            while (processed != max_avail && *(input_orig+processed) == ' ') {
                processed++;
            }
        }
    }

    *outputbuffer = 0; // "close" string
    return processed;
}


asm("
.data
.xdef help
help:
.even

      .word string0  - help
      .word string1  - help
      .word string2  - help
      .word string3  - help
      .word string4  - help
      .word string5  - help
      .word string6  - help
      .word string7  - help
      .word string8  - help
      .word string9  - help
      .word string10 - help
      .word string11 - help
      .word string12 - help
      .word string13 - help
      .word string14 - help
      .word string15 - help
      .word string16 - help
      .word string17 - help
      .word string18 - help
      .word string19 - help
      .word string20 - help
      .word string21 - help

string0:
    .asciz \"[DOWN],[+]\"
string1:
    .asciz \"next page\"
"
#if defined(LARGE_FONT) && !defined(TI89)
"
string2:
    .asciz \"[2nd][DOWN/+]\"
string3:
    .asciz \"10 pages forward\"
"
#else
"
string2:
    .asciz \"[2nd][DOWN],[2nd][+]\"
string3:
    .asciz \"10 pages forward\"
"
#endif
"
string4:
    .asciz \"[UP],[-]\"
string5:
    .asciz \"previous page\"
"
#if defined(LARGE_FONT) && !defined(TI89)
"
string6:
    .asciz \"[2nd][UP/-]\"
string7:
    .asciz \"10 pages back\"
"
#else
"string6:
    .asciz \"[2nd][UP],[2nd][-]\"
string7:
    .asciz \"10 pages back\"
"
#endif
"
string8:
    .asciz \"[LEFT]\"
string9:
    .asciz \"first page\"
string10:
    .asciz \"[RIGHT]\"
string11:
    .asciz \"last page\"
string12:
    .asciz \"[ESC]\"
string13:
    .asciz \"back to selection menu\"
string14:
    .asciz \"[F1]\"
string15:
    .asciz \"this page\"
string16:
    .asciz \"[F5]\"
string17:
    .asciz \"boss key (quick exit)\"
string18:
    .asciz \"[0]-[9]\"
string19:
    .asciz \"goto bookmark\"
string20:
    .asciz \"[x]\"
string21:
    .asciz \"set bookmark\"
");


//=============================================================================
// Displays help page
//=============================================================================
static inline void ShowHelp(void) {
    LCD_BUFFER         b;

    LCD_save(b);
    ClrScr();
    DrawTitleBar("TICT eBook Reader - Control Keys");

    asm("
    movem.l %%d3-%%d4/%%a2-%%a4,-(%%sp)

    moveq #11-1,%%d3
    moveq #10,%%d4
    
    lea help(%%pc),%%a2
    movea.l %%a2,%%a4
    "
#if defined(LARGE_FONT) && !defined(TI89)
    "   
    lea BigFastStringXY(%%pc),%%a3
    "
#else
    "   
    lea FastStringXY(%%pc),%%a3
    "
#endif
    "
    0:
    movea.l %%a4,%%a0
    adda.w (%%a2)+,%%a0
    
    move.w %%d4,%%d1
    clr.w %%d0
    jsr (%%a3)

    movea.l %%a4,%%a0
    adda.w (%%a2)+,%%a0
    
    move.w %%d4,%%d1
    "
#if defined(LARGE_FONT) && !defined(TI89)
    "
    moveq #100,%%d0
    "
#else
    "    
    moveq #80,%%d0
    "
#endif
    "
    jsr (%%a3)
    "
#if defined(LARGE_FONT) && !defined(TI89)
    "
    add.w #9,%%d4
    "
#else
    "
    addq.w #8,%%d4
    "
#endif
    "
    dbf %%d3,0b
    movem.l (%%sp)+,%%d3-%%d4/%%a2-%%a4

    " : : : "d0","d1","d2","a0","a1","cc","memory");
    GetUserInput();
    LCD_restore(b);
}



//=============================================================================
// handles display of a complete part
//=============================================================================
void DisplayPart(unsigned char* data, unsigned short length) {
    unsigned short offset = 0;
    unsigned short pages  = 0;
    short          line;
    char           tmp_str[100];
    short          input;
    unsigned short actpage;
    unsigned short pageoffset[MAX_PAGES]; // dirty hack - static offset table
    short          validinput;


#if defined(TIME_MEASUREMENT)
    //-------------------------------------------------------------------------
    // start timer for clocks and enter the main loop
    //-------------------------------------------------------------------------
    OSFreeTimer(USER_TIMER);
    OSRegisterTimer(USER_TIMER,1000);
    OSTimerRestart(USER_TIMER);
#endif

    //-------------------------------------------------------------------------
    // first parse number of pages and setup pageoffset table so we can use
    // page-up ...
    //-------------------------------------------------------------------------
    while (offset < length) {
        pageoffset[pages++] = offset;
        for (line=0;line<LINES_PER_PAGE;line++) {
             offset += NextLine(data+offset,tmp_str,length-offset);
             if (offset >= length) break;
        }
        if (pages >= MAX_PAGES) return;
    }

#if defined(TIME_MEASUREMENT)
    {
       unsigned long measure_val = OSTimerCurVal(USER_TIMER);
       //----------------------------------------------------------------------
       // stop timer again, turn grayscales off, restore old interrupt handlers
       // again and free memory
       //----------------------------------------------------------------------
       OSFreeTimer(USER_TIMER);
       sprintf(tmp_str,"%lu tics",(1000-measure_val)*50);
       DrawStr(10,10,tmp_str,A_REPLACE);
       ngetchx();
    }
#endif

    if (load_page != -1) {
        if (!load_page) actpage  = pages;
        else            actpage  = load_page;

        if (actpage > pages) actpage = pages;

        offset    = pageoffset[actpage-1];
        load_part = -1;
        load_page = -1;
    }
    else {
        actpage = 1;
        offset  = 0;
    }

    do {
        if (offset < length) {
            restore_status = 0;
            screenbuffer = doublebuffer;
            memset(doublebuffer,0,LCD_SIZE);

            FillFullLines(0,7,0xff);
            sprintf(tmp_str,"Part %u: page %u of %u",act_part,actpage,pages);

            FastStringXY(0,1,tmp_str);
            FastStringXY(WIDTH_IN_PIXELS-45,1,"[F1] for help");

            //-----------------------------------------------------------------
            // draw complete page
            //-----------------------------------------------------------------
            for (line=1;line<LINES_PER_PAGE+1;line++) {
                offset += NextLine(data+offset,tmp_str,length-offset);
#if defined(LARGE_FONT)
                BigFastStringXY(0,line*8,tmp_str);
#else
                FastStringXY(0,2+line*7,tmp_str);
#endif
                if (offset >= length) break;
            }
            screenbuffer = LCD_MEM;
            memcpy(screenbuffer,doublebuffer,LCD_SIZE);
            restore_status = 1;
        }

        validinput = 0;

        do {
            input = GetUserInput();

            //-----------------------------------------------------------------
            // NOTE: the if/else chaining is smaller than the
            //       equivalent switch statement !!!
            //-----------------------------------------------------------------
            if (input == K_UP || input == '-') {
                if (actpage > 1) {
                    offset = pageoffset[actpage-2];
                    actpage--;
                    validinput = 1;
                }
                else {
                    if (act_part > 1) {
                        load_part = act_part - 1;
                        load_page = 0;  // indicates last page !!!!
                        input = K_ESCAPE;
                    }
                }
            }
            //-----------------------------------------------------------------
            // [2nd]+[up]  or  [2nd]+[-]  ... 10 pages back
            //-----------------------------------------------------------------
            else if ((input == K_UP + K_SECOND) || (input == '-' + K_SECOND)) {
                if (actpage > 10) {
                    offset = pageoffset[actpage-11];
                    actpage -= 10;
                    validinput = 1;
                }
                else {
                    if (actpage != 1) {
                        offset     = 0;
                        actpage    = 1;
                        validinput = 1;
                    }
                }
            }
            else if (input == K_DOWN || input == '+') {
                if (actpage < pages) {
                    actpage++;
                    validinput = 1;
                }
                else {
                    if (act_part < nr_parts) {
                        load_part = act_part + 1;
                        load_page = 1;
                        input     = K_ESCAPE;
                    }
                }
            }
            //-----------------------------------------------------------------
            // [2nd]+[down]  or  [2nd]+[+]  ... 10 pages forward
            //-----------------------------------------------------------------
            else if ((input == K_DOWN + K_SECOND) || (input == '+' + K_SECOND)) {
                if (actpage + 10 <= pages) {
                    actpage += 10;
                    offset = pageoffset[actpage-1];
                    validinput = 1;
                }
                else {
                    if (actpage != pages) {
                        offset     = pageoffset[pages-1];
                        actpage    = pages;
                        validinput = 1;
                    }
                }
            }
            else if (input == K_LEFT) {
                if (actpage != 1) {
                    offset     = 0;
                    actpage    = 1;
                    validinput = 1;
                }
            }
            else if (input == K_RIGHT) {
                if (actpage != pages) {
                    offset     = pageoffset[pages-1];
                    actpage    = pages;
                    validinput = 1;
                }
            }
            else if (input == K_F1) {
                restore_status = 0;
                ShowHelp();
                restore_status = 1;
            }
            else if (input == 'x' || input == 'X') {
                // for the BookmarkDialog it is not necessary to
                // turn off status line restoring, because the
                // bookmark dialog will not "overdraw" the status
                // line area
                short ret = BookMarkDialog();

                if (ret != -1) {
                    unsigned short *ptr = &active_marks->part[0];
                    short i;
                    for (i=0;i<NR_BOOKMARKS;i++) {
                        *ptr = REAL_MARK_VALUE(*ptr);
                        ptr++;
                    }
                    active_marks->page[ret] = actpage;
                    active_marks->part[ret] = act_part | BOOKMARKS_MASK_LATEST;
                    SaveBookMarks();
                }
            }
            else if (isdigit(input)) {
                short num      = input - '0';
                short tmp_part = REAL_MARK_VALUE(active_marks->part[num]);

                // just do it when mark is okay ....
                if (active_marks->page[num]) {
                    if (tmp_part == act_part) {
                        actpage = active_marks->page[num];
                        if (actpage > pages) actpage = pages;
                        offset = pageoffset[actpage-1];
                        validinput = 1;
                    }
                    else {
                        if (tmp_part && tmp_part <= nr_parts) {
                            load_part = tmp_part;
                            load_page = active_marks->page[num];
                        }
                        input = K_ESCAPE;
                    }
                }
            }
        }
        while (!validinput && input != K_ESCAPE);
    }
    while (input != K_ESCAPE);
    restore_status = 0;
}



//=============================================================================
// handles part selection menu
//
// returns part index to view or -1 for escape
//=============================================================================
static inline short HandlePartSelection(void) {
    char  s[100];
    short input;

    ClrScr();
    DrawTitleBar(MENU_TOP);
    FillFullLines(HEIGHT_IN_PIXELS - 7,7,0xff);
    FastStringXY(C89_92(19,59),HEIGHT_IN_PIXELS-6,MENU_BOTTOM);

    memcpy(doublebuffer,LCD_MEM,LCD_SIZE);
    restore_status = 1;
    do {
        if (load_part != -1) {
            act_part = load_part;
            break;
        }
        sprintf(s,"Part %d of %d",act_part,nr_parts);
        FillFullLines(HEIGHT_IN_PIXELS/2-10,20,0);
        DrawStr(CalcCenterPosition(s,F_6x8),HEIGHT_IN_PIXELS/2-4,s,A_REPLACE);
        input = GetUserInput();
        if (input == K_LEFT || input == K_UP) {
            act_part--;
            act_part=(act_part+nr_parts-1)%nr_parts+1;  // warp around treatment
        }
        else if (input == K_RIGHT || input == K_DOWN) {
            act_part++;
            act_part=(act_part+nr_parts-1)%nr_parts+1;  // warp around treatment
        }
        else if (input == K_ESCAPE) {
            act_part = -1;
            break;
        }
        else if (isdigit(input)) {
            short num = input - '0';
            //-----------------------------------------------------------------
            // just do bookmark loading when mark is okay (page must be != 0)
            //-----------------------------------------------------------------
            // Trying to hand-optimize this block turns into bigger code.
            if (active_marks->page[num]) {
                load_part = REAL_MARK_VALUE(active_marks->part[num]);
                load_page = active_marks->page[num];
            }
        }
    }
    while (input != K_ENTER);

    if (act_part != -1) {
        sprintf(s,"Please wait while unpacking Part %d ...",act_part);
        FillFullLines(HEIGHT_IN_PIXELS/2-10,20,0);
        FastStringXY(CalcCenterPosition(s,F_4x6),HEIGHT_IN_PIXELS/2-4,s);
    }

    restore_status = 0;
    return act_part;
}



//=============================================================================
// check if given symentry is a TTArchive and if yes, try to extract its
// description, otherwise use symentry name as description
//
// returns 1 if okay,    0 otherwise
//=============================================================================
static inline short CheckArchiveAndGetDesc(SYM_ENTRY* symptr,unsigned char* buffer) {
    unsigned short* src;
    unsigned short size;
    unsigned char* comment;
    short          was_locked = HeapGetLock(symptr->handle);

    src = (unsigned short *)HLock(symptr->handle); // lock in any case even if it was already locked

    if (!src) return 0;

    // Handles start at an even address.
    size = *src++; //(((unsigned short)*src) << 8) + (unsigned short)*(src+1);

    if (!ttarchive_valid((void *)src)) {
        if (!was_locked) HeapUnlock(symptr->handle);
        return 0;
    }

    comment = ttarchive_info((void *)src);

    if (comment) strncpy(buffer,comment,TTARCHIVE_INFOLENGTH);
    else         strncpy(buffer,symptr->name,8);

    if (!was_locked) HeapUnlock(symptr->handle);
    return 1;
}



//=============================================================================
// book selection menu
//
// credits for this routine goes to Zeljko Juric !!
// (I have adapted this code from his ScottFree driver)
//=============================================================================
static inline HANDLE HandleBookSelection(short* lastbook) {
    short             select  = 0;
    short             index   = 1;
    short             found   = 0;
    short             i;
    short             handle  = PopupNew((char*)"SELECT AN eBOOK",C89_92(53,69));
    SYM_ENTRY*        symptr  = SymFindFirst(BOOKS_FOLDER,1);
    char              desc[TTARCHIVE_INFOLENGTH+1];

    ClrScr();
    DrawClipRect(ScrToWin(ScrRect),ScrRect,A_NORMAL);
    FastStringXY(C89_92(11,51),C89_92(85,110),"\xA9 TI-Chess Team   http://tict.ticalc.org");
    DrawStr(C89_92(11,51),3,READER_TITLE,A_NORMAL);

    while (symptr) {
        if (CheckArchiveAndGetDesc(symptr,desc)) {
            found++;
            PopupAddText(handle,-1,desc,index);
        }
        symptr=SymFindNext();
        index++;
    }

    if (found) select = PopupDo(handle,CENTER,CENTER,*lastbook);
    else {
        DrawStr(C89_92(26,66),30,"No eBooks found !!!",A_REPLACE);
        select = 0;
        GetUserInput();
    }
    ClrScr();
    HeapFree(handle);

    if (!select) return 0;

    *lastbook = select;
    symptr=SymFindFirst(BOOKS_FOLDER,1);

    for(i=1;i<select;i++) symptr=SymFindNext();

    active_book[8] = 0;
    strncpy(active_book,symptr->name,8);

    return symptr->handle;
}



//=============================================================================
// clears screen and outputs error message
//=============================================================================
void OutError(const char* s) {
    ClrScr();
    DrawStr(0,0,s,A_REPLACE);
    GetUserInput();
}



//=============================================================================
// handles display of book
//
// when this function is called we can be sure that its an TTArchive !!
//=============================================================================
void HandleBookReading(HANDLE h) {
    unsigned short* src;
    unsigned short length;
    unsigned char* data;
    short          i;
    short          was_locked = HeapGetLock(h);

    src = (unsigned short *)HLock(h); // lock in any case! even if it was locked previously

    if (!src) {
        OutError("ERROR: cannot lock book");
        return;
    }

    // Handles start at an even address.
    src++;

    i = ShowTitle((void *)src);

    if (!i || i == K_ESCAPE) {
        if (!was_locked) HeapUnlock(h);
        if (!i) {
            if (out_of_mem) OutError("ERROR: out of memory");
            else            OutError("ERROR: title not found");
        }
        return;
    }

    nr_parts = ttarchive_entries((void *)src) - 1;  // -1 is necessary due to the title picture

    //-------------------------------------------------------------------------
    // load bookmarks. nr_parts must be set previously. will set
    // load_part/load_page
    //-------------------------------------------------------------------------
    LoadBookMarks();
    act_part = 1;

    do {
        if ((HandlePartSelection()) == -1) break;

        if (!(data = GetDecompressedEntry((void *)src,act_part,&length))) {
            if (out_of_mem) OutError("ERROR: out of memory");
            else            OutError("ERROR: part not found");
            break;
        }
        DisplayPart(data,length);
        free(data);
        if (exit_completely) break;
    }
    while (1);

    if (!was_locked) HeapUnlock(h);
}



//=============================================================================
// nomen omen est
//=============================================================================
void _main(void) {
    SYM_ENTRY* symptr;
    BookMarks  tmpmarks;
    short      lastbook = 1;
    ESI        argptr;
    void*      orig_screen;
    HANDLE     h;
    short      previous_font = FontGetSys();

    //-------------------------------------------------------------------------
    // allocate space for the charset, width table and screenbuffers ...
    //-------------------------------------------------------------------------
    if (!(doublebuffer = malloc(LCD_SIZE*2))) {
        OutError("ERROR: out of memory");
        return;
    }

    orig_screen = doublebuffer + LCD_SIZE;

    //-------------------------------------------------------------------------
    // install statusline restoring interrupt handler
    //-------------------------------------------------------------------------
    old_inthandler = GetIntVec(AUTO_INT_5);
    restore_status = 0;
    SetIntVec(AUTO_INT_5,RestoreStatusHandler);

    //-------------------------------------------------------------------------
    // setup globals
    //-------------------------------------------------------------------------
    exit_completely  = 0;
    active_marks     = &tmpmarks;
    out_of_mem       = 0;

    SetupCharSet();
    FontSetSys(F_6x8);
    screenbuffer = LCD_MEM;

    LCD_save(orig_screen);

    //-------------------------------------------------------------------------
    // get commandline argument (name of variable to edit)
    //-------------------------------------------------------------------------
    InitArgPtr(argptr);

    if (GetArgType(argptr) == STR_TAG) {
        const char *fname = GetStrnArg(argptr);
        char tmpstr[40];

        tmpstr[0] = 0;

        if ((symptr = DerefSym(SymFind(strcpy(tmpstr + 1, fname) + strlen(fname))))) {
            active_book[8] = 0;
            strncpy(active_book,symptr->name,8);
            HandleBookReading(symptr->handle);
        }
    }
    else {
        do {
            //-----------------------------------------------------------------
            // open book selection menu
            //-----------------------------------------------------------------
            h = HandleBookSelection(&lastbook);
            if (!h) break;
            HandleBookReading(h);
            if (exit_completely) break;
        }
        while (1);
    }

    SetIntVec(AUTO_INT_5,old_inthandler);
    LCD_restore(orig_screen);
    free(doublebuffer);
    FontSetSys(previous_font);
}


//#############################################################################
//###################### NO MORE FAKES BEYOND THIS LINE #######################
//#############################################################################
//
//=============================================================================
// Revision History
//=============================================================================
//
// $Log: ebook.c,v $
// Revision 2.06  2004/07/14 14:31:55  DEBROUX Lionel
// recompiled with TIGCC 0.95 Beta 12
//
// Revision 2.05  2004/07/07 15:27:00  DEBROUX Lionel
// (1) cleaned up the source
// (2) fixed a bug with bookmarks
//
// Revision 2.04  2004/06/28 13:12:53  DEBROUX Lionel
// (1) suppressed that silly infinite loop in BookMarkDialog. Sorry for the
// inconvenience !
// (2) since -mpcrel is used, there are no relocations. F-Line relocations
// are therefore not useless, all the more they require a bigger internal
// F-Line instructions handler !
// (3) removed compilation option -ftracer, saving some more space (now
// 6689 bytes for the normal 89 version)
//
// Revision 2.03  2004/06/28 13:11:27  DEBROUX Lionel
// (1) recompilation with TIGCC 0.95
// (2) code optimizations
// (3) now using F-Line ROM_CALLs
// (4) asm("trap #4"); changed back to off(); (no impact on size thanks to
// F-Line ROM_CALLs)
// (5) inclued code to save power in grayscale mode
//
// Revision 2.02  2002/11/21 12:27:40  nussbaumer
// changes due to TIGCC 0.94
//
// Revision 2.01  2002/09/20 12:31:53  tnussb
// modifications for "large" font version for TI89
//
// Revision 2.00  2002/09/20 09:30:03  tnussb
// forced version to 2.00
//
// Revision 1.94  2002/09/20 09:26:49  tnussb
// (1) leading spaces after a linebreak will not removed anymore
// (2) large font support for TI92p/V200 when compiling with -DLARGE_FONT
//     (thanx to janjan2 for this modification)
// (3) title screen will be scaled up if compiling with -DLARGE_FONT (TI92p/V200)
//     (thanx to janjan2 for this modification)
//
// Revision 1.93  2002/08/01 12:08:05  tnussb
// (1) off() changed to "trap #4" (thanx to Joel Thompson - joelt49@yahoo.com)
// (2) 10 pages forward/back by holding [2nd] while paging (thanx to Joel again)
// (3) help page modified for (2)
//
// Revision 1.92  2002/05/23 06:29:46  tnussb
// version raised to reflect the use of the old unpacking version
//
// Revision 1.91  2002/03/28 21:14:32  tnussb
// again size optimizations by Francesco Orabona (bremen79@infinito.it)
//
// Revision 1.90  2002/03/13 15:07:35  tnussb
// no changes - just the version number raised, because a new exepack decompression
// function is used
//
// Revision 1.89  2002/03/06 16:13:29  tnussb
// minor size optimizations by Francesco Orabona (bremen79@infinito.it)
//
// Revision 1.88  2002/03/04 08:58:05  tnussb
// don't disable statusline restoring during bookmark dialog is shown (not necessary)
//
// Revision 1.87  2002/03/01 17:14:52  tnussb
// Restores now statusline area by using an own INT5 handler. This wastes
// again 400 bytes, but I think thats okay.
//
// Revision 1.86  2002/03/01 16:32:46  tnussb
// (1) Thanx to Austin Chu (qarv@yahoo.com) for pointing me out how stupid
//     and therefore slow NextLine() was implemented previously
// (2) many parts rewritten to optimize the size (now the size is 7317 Bytes
//     on the TI89 - great, much space to waste in further versions ;-)
// (3) "white on black" text is now drawn with FastStringXY, too
// (4) uses now FillFullLines macro instead of RectFill()
// (5) previously set font will be restored now before exit
// (6) FontSetSys() only used in CharSet grabbing, because all 4x6 strings
//     are now drawn by FastStringXY
// (7) I have moved the bottom line to the top, because I think this way
//     reading the text is a little bit easier. If the user presses now any
//     key which turns on a status flag, he has to refresh the screen by
//     his own by pressing once pageup+pagedown. This problem should be fixed
//     in one of the next versions.
//
// Revision 1.85  2002/02/27 11:16:44  tnussb
// 328 bytes saved by declaring some functions "static inline" and
// by using "static const" for some character arrays
//
// Revision 1.84  2002/02/13 18:13:02  tnussb
// (1) previously broken sprite drawing fixed again
// (2) FastDrawStringXY() heavily optimized
// (3) allocating now also memory for doublebuffer
//
// Revision 1.83  2002/02/11 09:43:24  tnussb
// (1) the buffer used to save screen and the charwidth are no
//     more on the stack, to avoid stack overflows (Lionel Debroux)
// (2) minor size optimizations
//
// Revision 1.82  2002/02/11 08:35:47  tnussb
// all warnings reported by compiler option -Wwrite-strings fixed
//
// Revision 1.81  2002/02/07 10:00:15  tnussb
// (1) uses now NO_EXIT_SUPPORT to reduce code size
// (2) minor fix to suppress a compiler warning of TIGCC 0.93
//
// Revision 1.80  2001/08/01 21:18:14  Thomas Nussbaumer
// (1) serious bug in bookmark dialog handling fixed (previously
//     not wanted keys had escaped from the routine, too)
// (2) [+] and [+] key can be used for paging
//
// Revision 1.77  2001/06/20 13:00:39  Thomas Nussbaumer
// now you can press 'x' twice instead of 'x' + '0'
//
// Revision 1.76  2001/04/08 21:04:05  Thomas Nussbaumer
// (1) using now idle() in GetUserInput()
// (2) letter 'g' displayed now one pixel lower than the others
//
// Revision 1.75  2001/03/21 21:41:03  Thomas Nussbaumer
// changed version date and version number forced to 1.75
//
// Revision 1.71  2001/02/04 16:45:13  Thomas Nussbaumer
// (1) using 7 pixels per line for better readability
// (2) using new header files for decompression and ttarchive access
//
// Revision 1.70  2001/01/22 23:09:33  Thomas Nussbaumer
// (1) using now again complete character set (0..255), but grabbed from a
// backbuffer during startup
// (2) stating now out-of-memory instead of part-not-found or title-not-found
// (3) broken bookmark feature fixed
// (4) using now HANDLEs wherever possible and applicable instead of SYM_ENTRYs
//
// Revision 1.63  2001/01/16 02:11:47  Thomas Nussbaumer
// fixed problem when reader is in tictebks directory again
//
// Revision 1.62  2001/01/16 01:12:48  Thomas Nussbaumer
// now it should be save to install the reader in the tictebks folder, too
//
// Revision 1.61  2001/01/10 22:15:05  Thomas Nussbaumer
// (1) own font 4x6 drawing routine to speed up page drawing added
// (2) some size optimizations performed to keep size under 8kB
//
// Revision 1.60  2001/01/06 10:47:07  Thomas Nussbaumer
// (1) problems reported by compiling with -Wall fixed
// (2) necessary changes for TIGCC Library 2.31 (int<->short)
// (3) can handle now commandline argument
//
// Revision 1.57  2000/10/01 14:55:57  Thomas Nussbaumer
// problem with 'w' & 'm' characters fixed
//
// Revision 1.56  2000/08/30 20:01:55  Thomas Nussbaumer
// using now assembler optimized version of unpack routine
//
// Revision 1.55  2000/08/30 19:41:19  Thomas Nussbaumer
// demonstration of nostub library usage (#define LIBRARY_VERSION) added
//
// Revision 1.54  2000/08/29 12:52:05  Thomas Nussbaumer
// saved 80 bytes more. Due to the optimized decompression algorithm
// we have about 800 bytes free for further improvements.
// I wonder if the upcoming metatag handling can fit into 800 bytes.
// The rendering engine will undergo heavy reconstructions for
// supporting the variable alignment of the text.
//
// Revision 1.53  2000/08/27 23:43:07  Thomas Nussbaumer
// (1) if there is no latest saved bookmark automatically load page1 of part1
// (2) try to load previous/next part only if not on first or last part already
//
// Revision 1.52  2000/08/27 20:55:08  Thomas Nussbaumer
// (1) reloads now at latest saved bookmark
// (2) on the end of a part you can now page further down to go to the next
// part
//
// Revision 1.51  2000/08/27 19:06:49  Thomas Nussbaumer
// (1) size reduced about more than 600 bytes by using now
//     separate builds instead of runtime calculator queries
// (2) speedup of line length calculation by doing it by hand
//     instead using function CharWidth()
//
// Revision 1.50  2000/08/23 20:35:48  Thomas Nussbaumer
// forcing version 1.50 (official released)
//
// Revision 1.4  2000/08/23 01:20:56  Thomas Nussbaumer
// (1) handling of 20 characters eBook description added
// (2) show only valid ttarchives in booklist
// (3) bug corrected in FindBookMarks(). this bug has
//     crashed my poor TI92p !!! that was the first crash
//     since about 6 month. normally my software will not
//     do such things ;-)
// (4) heavily reduced again to fit in 8kB. now we have
//     still 181 bytes left for further extensions
//
// Revision 1.3  2000/08/21 23:50:11  Thomas Nussbaumer
// (1) handling now APD (auto-power-down)
// (2) handling of calculator OFF key combinations
// (3) books must be stored now in folder tictebks (for autofind)
// (4) using now book selection box instead of commandline argument
// (5) up to 10 bookmarks can be set for each book
// (6) bookmarks are stored in folder tictmark
// (7) jumping to a bookmark is done by pressing one of the
//     numerical key (0-9)
// (8) book mark setting is done by pressing 'x' and a numerical key
// (9) fits exactly in 8kB (AMS 2.03) - just 7 bytes left ;-)
//
// Revision 1.2  2000/08/20 23:01:57  Thomas Nussbaumer
// missing free() added in ShowTitle()
//
// Revision 1.1  2000/08/20 16:51:19  Thomas Nussbaumer
// initial version
//
